var _ = require('underscore'),
    ns;

/**
 Nests new attributes by using namespace string and value in a provided object
 Example:
 ns({}, 'Foo.1.bar', true); // returns Object {Foo: {1: {bar: true}}}

 @function ns
 @param {Object} context Object that we want to extend
 @param {String} nspace Namespace string
 @param {Mixed} payload Value of this namespace
 @return {Object} Returns Original object with added namespace
 */

ns = function (context, nspace, payload) {
    var iter = function (obj, parts, value) {
        var part = parts.shift();

        obj[part] = obj[part] || {};

        if (parts.length === 0) {
            obj[part] = value;
            return obj;
        } else {
            return iter(obj[part], parts, value);
        }
    };

    iter(context, nspace.split('.'), payload);

    return context;
}

/**
 Takes a nested object and returns value of a namespaced string if it exists in this object
 Example:
 ns.get({Foo: {1: {bar: true}}}, 'Foo'); // returns Object {1: {bar: true}}
 ns.get({Foo: {1: {bar: true}}}, 'Foo.1.bar'); // returns true

 @method get
 @static
 @param {Object} context Object in which we will look for namespace value
 @param {String} nspace Namespace string which value we're looking for
 @return {Mixed} Returns value of the attribute or undefined
 */

ns.get = function (context, nspace) {
    var result = _.reduce(nspace.split('.'), function (memo, part) {
        if (memo !== undefined) {
            return memo[part];
        } else {
            return undefined;
        }
    }, context);

    return result;
}

/**
 Converts a nested object info flat object where keys are namespaces
 Example:
 ns.stringify({Foo: {1: {bar: true}}}); // returns Object {Foo.1.bar: true}
 ns.stringify({Foo: {1: {bar: true, baz: false}}}); // returns Object {Foo.1.bar: true, Foo.1.baz: false}

 @method stringify
 @static
 @param {Object} context Object that needs to be converted
 @return {Object} Returns namespaced object
 */

ns.stringify = function (context) {
    var payload = {};

    var iter = function (context, str) {
        str = str || '';

        for (var prop in context) {
            if (context.hasOwnProperty(prop)) {

                var new_str = (str.length > 0) ? str + '.' + prop : prop;

                if (typeof context[prop] === 'object') {
                    iter(context[prop], new_str);
                } else {
                    payload[new_str] = context[prop];
                }
            }
        }
    };

    iter(context);
    return payload;
}

module.exports = ns;